import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer_plugin/protocol/protocol_common.dart';
import 'package:custom_flutter_analyze/logger/log.dart';
import 'package:custom_flutter_analyze/utils/utils.dart';
import 'package:custom_flutter_analyze/visitors/forbidden_method_visitor.dart';
import 'package:glob/glob.dart';
import 'package:path/path.dart';
import 'package:yaml/yaml.dart';

List<String> excludedFilesFromAnalysisOptions(File analysisOptions) {
  final parsedOptions = loadYaml(analysisOptions.readAsStringSync()) as YamlMap;
  final analyzerSection = parsedOptions.nodes['analyzer'];
  if (analysisOptions != null) {
    final dynamic excludedSection = (analyzerSection as YamlMap)['exclude'];
    if (excludedSection != null) {
      // ignore: avoid_annotating_with_dynamic
      return (excludedSection as YamlList).map((dynamic path) => path as String).toList();
    }
  }
  return [];
}

List<String> resolvePaths(List<String> paths, List<String> excludedFolders) {
  final excludedGlobs = excludedFolders.map((path) => Glob(path)).toList();
  return paths
      .expand((path) => Glob('$path/**/*.dart').listSync().whereType<File>().where((file) => !_isExcluded(file.path, excludedGlobs)).map((e) => e.path))
      .toList();
}

bool _isExcluded(String filePath, Iterable<Glob> excludes) => excludes.any((exclude) => exclude.matches(filePath));

var stopWatch = Stopwatch();
Future<List<AnalysisError>> collectAnalyzerErrors(AnalysisContextCollection analysisContextCollection, List<String> paths) async {
  final analysisErrors = <AnalysisError>[];
  for (final filePath in paths) {
    final normalizedPath = normalize(filePath);
    stopWatch.start();
    ResolvedUnitResult resolvedUnitResult = await analysisContextCollection.contextFor(normalizedPath).currentSession.getResolvedUnit(normalizedPath);
    stopWatch.stop();
    log.info('resolvedUnitResult timeCost = ${stopWatch.elapsedMilliseconds}');

    ForbiddenMethodVisitor forbiddenMethodVisitor = ForbiddenMethodVisitor(analysisResult: resolvedUnitResult, methodName: 'forbiddenMethod', errorCode: '0');
    resolvedUnitResult.unit.accept(forbiddenMethodVisitor);
    log.info('forbiddenMethodVisitor.nodes.length=${forbiddenMethodVisitor.nodes.length}');
    analysisErrors.addAll(forbiddenMethodVisitor.nodes.map((e) => getAnalysisError(astNode: e, analysisResult: resolvedUnitResult)));

    // ParsedUnitResult parsedUnit = await analysisContextCollection.contextFor(normalizedPath).currentSession.getParsedUnit(normalizedPath);
    // ForbiddenMethodVisitor2 forbiddenMethodVisitor2 = ForbiddenMethodVisitor2(analysisResult: parsedUnit, methodName: 'forbiddenMethod', errorCode: '0');
    // parsedUnit.unit.accept(forbiddenMethodVisitor2);
    // log.info('forbiddenMethodVisitor2.nodes.length=${forbiddenMethodVisitor2.nodes.length}');
  }
  log.info('analysisErrors ===$analysisErrors');
  return analysisErrors;
}

String readableAnalysisError(AnalysisError analysisError) => analysisError.toReadableString();

AnalysisError codeIssueToAnalysisError(Issue issue, ResolvedUnitResult analysisResult) {
  return AnalysisError(
    issue.errorSeverity,
    issue.errorType,
    Location(
      analysisResult.unit.declaredElement.source.fullName,
      issue.offset,
      issue.length,
      issue.line,
      issue.column,
    ),
    issue.message,
    issue.code,
    correction: issue.correction,
    hasFix: issue.hasFix,
  );
}

extension ReadableOutput on AnalysisError {
  String toReadableString() => '$severity - $type\n$message\n${location.file}:${location.startLine}:${location.startColumn}';
}

class Issue {
  Issue({
    this.hasFix,
    this.offset,
    this.length,
    this.line,
    this.column,
    this.message,
    this.code,
    this.correction,
    this.comment,
    this.replacement,
    this.errorSeverity,
    this.errorType,
  });

  final bool hasFix;

  final int offset;
  final int length;
  final int line;
  final int column;
  final String message;
  final String code;
  final String correction;
  final String comment;
  final String replacement;

  final AnalysisErrorSeverity errorSeverity;
  final AnalysisErrorType errorType;

  @override
  String toString() {
    return 'Issue{hasFix: $hasFix, offset: $offset, length: $length, line: $line, column: $column, message: $message, code: $code, correction: $correction, comment: $comment, replacement: $replacement, errorSeverity: $errorSeverity, errorType: $errorType}';
  }
}
