import 'dart:async';

import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/file_system/file_system.dart';
import 'package:analyzer/source/line_info.dart';
import 'package:analyzer/src/context/builder.dart';
import 'package:analyzer/src/context/context_root.dart' as analyzer;
import 'package:analyzer/src/dart/analysis/driver.dart';
import 'package:analyzer_plugin/plugin/plugin.dart';
import 'package:analyzer_plugin/protocol/protocol_common.dart';
import 'package:analyzer_plugin/protocol/protocol_generated.dart' as plugin;
import 'package:glob/glob.dart';
import 'package:pub_semver/pub_semver.dart';

import 'logger/log.dart';
import 'utils/utils.dart';
import 'visitors/forbidden_method_visitor.dart';

class MirrorPlugin extends ServerPlugin {
  MirrorPlugin(ResourceProvider provider) : super(provider);

  var _filesFromSetPriorityFilesRequest = <String>[];
  String rootPath;

  @override
  List<String> get fileGlobsToAnalyze => const ['*.dart'];

  Iterable<Glob> excludeFileGlobs;

  @override
  String get name => 'Custom Analyzer Plugin';

  @override
  String get version => '0.0.1';

  @override
  bool isCompatibleWith(Version serverVersion) => true;

  @override
  void contentChanged(String path) {
    log.info("contentChanged$path");
    AnalysisDriverGeneric driver = super.driverForPath(path);
    driver.addFile(path);
  }

  @override
  AnalysisDriverGeneric createAnalysisDriver(plugin.ContextRoot contextRoot) {
    final analysisRoot = analyzer.ContextRoot(contextRoot.root, contextRoot.exclude, pathContext: resourceProvider.pathContext)
      ..optionsFilePath = contextRoot.optionsFile;

    final contextBuilder = ContextBuilder(resourceProvider, sdkManager, null)
      ..analysisDriverScheduler = analysisDriverScheduler
      ..byteStore = byteStore
      ..performanceLog = performanceLog
      ..fileContentOverlay = fileContentOverlay;

    final dartDriver = contextBuilder.buildDriver(analysisRoot);
    rootPath = contextRoot.root;

    runZonedGuarded(() async {
      dartDriver.results.listen((analysisResult) {
        _processResult(dartDriver, analysisResult);
      });
    }, (e, stackTrace) {
      channel.sendNotification(plugin.PluginErrorParams(false, e.toString(), stackTrace.toString()).toNotification());
    });
    return dartDriver;
  }

  // @override
  // Future<plugin.EditGetFixesResult> handleEditGetFixes(plugin.EditGetFixesParams parameters) async {
  //   try {
  //     final driver = driverForPath(parameters.file) as AnalysisDriver;
  //     final analysisResult = await driver.getResult(parameters.file);
  //
  //     final fixes = _check(driver, analysisResult)
  //         .where((fix) =>
  //             fix.error.location.file == parameters.file &&
  //             fix.error.location.offset <= parameters.offset &&
  //             parameters.offset <= fix.error.location.offset + fix.error.location.length &&
  //             fix.fixes.isNotEmpty)
  //         .toList();
  //
  //     return plugin.EditGetFixesResult(fixes);
  //   } on Exception catch (e, stackTrace) {
  //     channel.sendNotification(plugin.PluginErrorParams(false, e.toString(), stackTrace.toString()).toNotification());
  //
  //     return plugin.EditGetFixesResult([]);
  //   }
  // }

  @override
  Future<plugin.AnalysisSetPriorityFilesResult> handleAnalysisSetPriorityFiles(plugin.AnalysisSetPriorityFilesParams parameters) async {
    _filesFromSetPriorityFilesRequest = parameters.files;
    _updatePriorityFiles();
    return plugin.AnalysisSetPriorityFilesResult();
  }

  void _updatePriorityFiles() {
    final filesToFullyResolve = {
      ..._filesFromSetPriorityFilesRequest,
      for (final driver2 in driverMap.values) ...(driver2 as AnalysisDriver).addedFiles,
    };
    final filesByDriver = <AnalysisDriverGeneric, List<String>>{};
    for (final file in filesToFullyResolve) {
      final contextRoot = contextRootContaining(file);
      if (contextRoot != null) {
        final driver = driverMap[contextRoot];
        filesByDriver.putIfAbsent(driver, () => <String>[]).add(file);
      }
    }
    filesByDriver.forEach((driver, files) => driver.priorityFiles = files);
  }

  void _processResult(AnalysisDriver driver, ResolvedUnitResult analysisResult) {
    log.info('_processResult : ${analysisResult.path}');
    try {
      final analysisErrors = <AnalysisError>[];
      if (analysisResult.unit != null && analysisResult.libraryElement != null) {
        ForbiddenMethodVisitor forbiddenMethodVisitor = ForbiddenMethodVisitor(analysisResult: analysisResult, methodName: 'forbiddenMethod', errorCode: '0');
        analysisResult.unit.accept(forbiddenMethodVisitor);
        log.info('forbiddenMethodVisitor.nodes.length=${forbiddenMethodVisitor.nodes.length}');
        analysisErrors.addAll(forbiddenMethodVisitor.nodes.map((e) => getAnalysisError(astNode: e, analysisResult: analysisResult)));

        channel.sendNotification(plugin.AnalysisErrorsParams(analysisResult.path, analysisErrors).toNotification());
      } else {
        channel.sendNotification(plugin.AnalysisErrorsParams(analysisResult.path, []).toNotification());
      }
    } on Exception catch (e, stackTrace) {
      channel.sendNotification(plugin.PluginErrorParams(false, e.toString(), stackTrace.toString()).toNotification());
    }
  }
}

List<int> getIgnoresLines(String content, LineInfo info, String ignoreId) {
  List<int> results = [];
  final _ignoreMatchers = RegExp('//[ ]*ignore:(.*)$ignoreId', multiLine: true);
  for (final match in _ignoreMatchers.allMatches(content)) {
    final location = info.getLocation(match.start);
    final lineNumber = location.lineNumber;
    results.add(lineNumber);
  }
  return results;
}
